from transforms.base import *
import struct
import math


class Deflate(Transform):
    """
    Deflate -> pack a stream
    window_size: base-two logarithm of window size, must be >= 8 and <= 15, defaults to 15
    """
    category = "compress"
    name = "deflate"
    icon = "wxART_FILETYPE_ARCHIVE"

    def run(self, data:bytes, window_size:int=15):
        import zlib
        if window_size < 8 or window_size > 15:
            raise ValueError("Invalid window_size value")
        o = zlib.compressobj(wbits=-window_size)
        return o.compress(data) + o.flush()

class Inflate(Transform):
    """
    Inflate -> unpack an deflated stream
    window_size: base-two logarithm of window size, must be >= 8 and <= 15, defaults to 15
    """
    category = "compress"
    name = "inflate"
    icon = "wxART_FILE_COMPRESSED"

    def run(self, data:bytes, window_size:int=15):
        import zlib
        if window_size < 8 or window_size > 15:
            raise ValueError("Invalid window_size value")
        return zlib.decompress(data, wbits=-window_size)


class ZlibCompress(Transform):
    """
    Compress a stream using zlib
    window_size: base-two logarithm of window size, must be >= 8 and <= 15, defaults to 15
    """
    category = "compress"
    name = "zlib compress"
    icon = "wxART_FILETYPE_ARCHIVE"

    def run(self, data:bytes, window_size:int=15):
        import zlib
        if window_size < 8 or window_size > 15:
            raise ValueError("Invalid window_size value")
        o = zlib.compressobj(wbits=window_size)
        return o.compress(data) + o.flush()

class ZlibDecompress(Transform):
    """
    Uncompress a zlib stream
    """
    category = "compress"
    name = "zlib decompress"
    icon = "wxART_FILE_COMPRESSED"

    def run(self, data:bytes):
        import zlib
        return zlib.decompress(data)    


class LzmaCompress(Transform):
    """
    Compress a stream using LZMA (with or without xz header)
    """
    category = "compress"
    name = "lzma compress"
    icon = "wxART_FILETYPE_ARCHIVE"

    def run(self, data:bytes, format:["XZ", "ALONE"]="ALONE", compression:int=6):
        import lzma
        if format == "ALONE":
            format = lzma.FORMAT_ALONE
        else:
            format = lzma.FORMAT_XZ
        return lzma.compress(data, format=format, preset=compression)

class LzmaDecompress(Transform):
    """
    uncompress a (raw) lzma stream
    """
    category = "compress"
    name = "lzma decompress"
    icon = "wxART_FILE_COMPRESSED"

    def run(self, data:bytes, raw_mode:bool=False, lzma2:bool=False, max_length:int=-1):
        import lzma
        if raw_mode:
            if lzma2:
                props = lzma._decode_filter_properties(lzma.FILTER_LZMA2, data[0:1])
                data = data[1:]
            else:
                props = lzma._decode_filter_properties(lzma.FILTER_LZMA1, data[0:5])
                data = data[5:]
            dec = lzma.LZMADecompressor(lzma.FORMAT_RAW, filters=[props])
        else:
            dec = lzma.LZMADecompressor()
        return dec.decompress(data, max_length=max_length)        


class Bzip2Compress(Transform):
    """
    Compress a stream using bzip2
    level: compression level between 1 and 9
    """
    category = "compress"
    name = "bz2 compress"
    icon = "wxART_FILETYPE_ARCHIVE"

    def run(self, data:bytes, level:int=9):
        import bz2
        if level < 1 or level > 9:
            raise ValueError("Invalid level value")
        return bz2.compress(data, level)


class Bzip2Decompress(Transform):
    """
    Decompress a stream using bzip2
    """
    category = "compress"
    name = "bz2 decompress"
    icon = "wxART_FILE_COMPRESSED"

    def run(self, data:bytes):
        import bz2
        return bz2.decompress(data)


class LzoCompress(Transform):
    """
    Compress using LZO
    level: compression level between 1 (default) and 9
    header: include metadata header for decompression in the output (default: True)
    """
    category = "compress"
    name = "lzo compress"
    icon = "wxART_FILETYPE_ARCHIVE"

    def run(self, data:bytes, level:int=1, header:bool=True):
        try:
            import lzo
        except ImportError:
            raise ImportError("You need to install python-lzo library first. On Windows I suggest that you use some pre-compiled wheel")
        if level < 1 or level > 9:
            raise ValueError("Invalid level value")
        return lzo.compress(data, level, header)

class LzoDecompress(Transform):
    """
    Decompress using LZO
    header: header is included in input (default: True)
    buflen: if header is False, a buffer length in bytes must be given that will fit the output
    """
    category = "compress"
    name = "lzo decompress"
    icon = "wxART_FILE_COMPRESSED"

    def run(self, data:bytes, header:bool=True, buflen:int=0):
        try:
            import lzo
        except ImportError:
            raise ImportError("You need to install python-lzo library first. On WIndows I suggest that you use some pre-compiled wheel")
        return lzo.decompress(data, header, buflen)            

class Lznt1Compress(Transform):
    """
    Compress using LZNT1 (RtlCompressBuffer)
    header: include metadata header for decompression in the output (default: True)
    """
    category = "compress"
    name = "lznt1 compress"
    icon = "wxART_FILETYPE_ARCHIVE"

    def run(self, data:bytes, chunk_size:int=1000):
        from .libs import lznt1
        return lznt1.compress(data, chunk_size)

class Lznt1Decompress(Transform):
    """
    Decompress using LZNT1 (RtlDecompressBuffer)
    length_check: proofcheck chunks length
    """
    category = "compress"
    name = "lznt1 decompress"
    icon = "wxART_FILE_COMPRESSED"

    def run(self, data:bytes, length_check:bool=True):
        from .libs import lznt1
        return lznt1.decompress(data, length_check)         


class OfficeRleUnpack(Transform):
    """
    RLE decoding as used in office documents
    """
    category = "compress"
    name = "Office RLE"
    icon = "wxART_FILE_COMPRESSED"

    def run(self, data:bytes):
        result = bytearray()
        if len(data) <= 1 or data[0] != 1:
            raise ValueError("Invalid data")
        ptr = 1
        while ptr < len(data):
            block_ptr = ptr
            header, = struct.unpack("<H", data[ptr:ptr+2])
            block_size = min((header & 0xFFF) + 3, len(data) - ptr)
            sig = (header >> 12) & 0x07
            if sig != 3 and False:
                raise ValueError("Invalid header signature at {}: 0x{:x}".format(ptr, sig))
            is_compressed = (header >> 15) != 0
            ptr += 2
            if not is_compressed:
                # uncompressed chunk
                result.extend(data[ptr:ptr+4096])
                ptr += 4096
            else:
                # comrpessed chunk
                dptr = len(result)
                while ptr < block_ptr + block_size and ptr < len(data):
                    flagbyte = data[ptr]
                    ptr += 1
                    for i in range(8):
                        if ptr >= block_ptr + block_size:
                            break
                        is_copy = (flagbyte & 1) == 1
                        flagbyte = flagbyte >> 1
                        if not is_copy:
                            # literal token
                            result.append(data[ptr])
                            ptr += 1
                        else:
                            # copy token
                            token, = struct.unpack("<H", data[ptr:ptr+2])
                            ptr += 2
                            bit_count = max(4, int(math.ceil(math.log(len(result) - dptr, 2))))
                            length_mask = 0xFFFF >> bit_count
                            offset_mask = ~length_mask
                            length = (token & length_mask) + 3
                            temp1 = token & offset_mask
                            temp2 = 16 - bit_count
                            offset = (temp1 >> temp2) + 1
                            start_point = len(result) - offset
                            for i in range(start_point, start_point + length):
                                result.append(result[i])
        return bytes(result)

